﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using Akka.Actor.Internals;
using Akka.Configuration;
using Akka.Dispatch;
using Akka.Dispatch.SysMsg;
using Akka.Event;
using Akka.Util;
using Debug = System.Diagnostics.Debug;

namespace Akka.Actor
{
    /// <summary>
    ///     An actor system is a hierarchical group of actors which share common
    ///     configuration, e.g. dispatchers, deployments, remote capabilities and
    ///     addresses. It is also the entry point for creating or looking up actors.
    ///     There are several possibilities for creating actors (see [[Akka.Actor.Props]]
    ///     for details on `props`):
    ///     <code>
    /// system.ActorOf(props, "name");
    /// system.ActorOf(props);
    /// system.ActorOf(Props.Create(typeof(MyActor)), "name");
    /// system.ActorOf(Props.Create(() =&gt; new MyActor(arg1, arg2), "name");
    /// </code>
    ///     Where no name is given explicitly, one will be automatically generated.
    ///     <b>
    ///         <i>Important Notice:</i>
    ///     </b>
    ///     This class is not meant to be extended by user code.
    /// </summary>
    public abstract class ActorSystem : ActorRefFactory, IDisposable
    {
        /// <summary>Gets the settings.</summary>
        /// <value>The settings.</value>
        public abstract Settings Settings { get; }

        /// <summary>Gets the name of this system.</summary>
        /// <value>The name.</value>
        public abstract string Name { get; }

        /// <summary>Gets the serialization.</summary>
        /// <value>The serialization.</value>
        public abstract Serialization.Serialization Serialization { get; }

        /// <summary>Gets the event stream.</summary>
        /// <value>The event stream.</value>
        public abstract EventStream EventStream { get; }

        /// <summary>
        ///     Gets the dead letters.
        /// </summary>
        /// <value>The dead letters.</value>
        public abstract ActorRef DeadLetters { get; }

        /// <summary>Gets the dispatchers.</summary>
        /// <value>The dispatchers.</value>
        public abstract Dispatchers Dispatchers { get; }

        /// <summary>Gets the mailboxes.</summary>
        /// <value>The mailboxes.</value>
        public abstract Mailboxes Mailboxes { get; }


        /// <summary>Gets the scheduler.</summary>
        /// <value>The scheduler.</value>
        public abstract Scheduler Scheduler { get; }

        /// <summary>Gets the log</summary>
        public abstract LoggingAdapter Log { get; }

        /// <summary>
        ///     Creates a new ActorSystem with the specified name, and the specified Config
        /// </summary>
        /// <param name="name">Name of the ActorSystem</param>
        /// <param name="config">Configuration of the ActorSystem</param>
        /// <returns>ActorSystem.</returns>
        public static ActorSystem Create(string name, Config config)
        {
            // var withFallback = config.WithFallback(ConfigurationFactory.Default());
            return CreateAndStartSystem(name, config);
        }

        /// <summary>
        ///     Creates the specified name.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns>ActorSystem.</returns>
        public static ActorSystem Create(string name)
        {
            return CreateAndStartSystem(name,ConfigurationFactory.Load());
        }

        private static ActorSystem CreateAndStartSystem(string name, Config withFallback)
        {
            var system = new ActorSystemImpl(name, withFallback);
            system.Start();
            return system;
        }

        /// <summary>
        /// Returns an extension registered to this ActorSystem
        /// </summary>
        public abstract object GetExtension(IExtensionId extensionId);

        /// <summary>
        /// Returns an extension registered to this ActorSystem
        /// </summary>
        public abstract T GetExtension<T>() where T : class, IExtension;

        /// <summary>
        /// Determines whether this instance has the specified extension.
        /// </summary>
        public abstract bool HasExtension(Type t);

        /// <summary>
        /// Determines whether this instance has the specified extension.
        /// </summary>
        public abstract bool HasExtension<T>() where T : class, IExtension;

        /// <summary>
        /// Tries to the get the extension of specified type.
        /// </summary>
        public abstract bool TryGetExtension(Type extensionType, out object extension);

        /// <summary>
        /// Tries to the get the extension of specified type.
        /// </summary>
        public abstract bool TryGetExtension<T>(out T extension) where T : class, IExtension;


        /// <summary>
        ///     Stop this actor system. This will stop the guardian actor, which in turn
        ///     will recursively stop all its child actors, then the system guardian
        ///     (below which the logging actors reside) and the execute all registered
        ///     termination handlers (<see cref="ActorSystem.RegisterOnTermination" />).
        /// </summary>
        public abstract void Shutdown();

        /// <summary>
        /// Returns a task that will be completed when the system has terminated.
        /// </summary>
        public abstract Task TerminationTask { get; }

        /// <summary>
        /// Block current thread until the system has been shutdown.
        /// This will block until after all on termination callbacks have been run.
        /// </summary>
        public abstract void AwaitTermination();

        /// <summary>
        /// Block current thread until the system has been shutdown, or the specified
        /// timeout has elapsed. 
        /// This will block until after all on termination callbacks have been run.
        /// <para>Returns <c>true</c> if the system was shutdown during the specified time;
        /// <c>false</c> if it timed out.</para>
        /// </summary>
        /// <param name="timeout">The timeout.</param>
        /// <returns>Returns <c>true</c> if the system was shutdown during the specified time;
        /// <c>false</c> if it timed out.</returns>
        public abstract bool AwaitTermination(TimeSpan timeout);

        /// <summary>
        /// Block current thread until the system has been shutdown, or the specified
        /// timeout has elapsed, or the cancellationToken was canceled. 
        /// This will block until after all on termination callbacks have been run.
        /// <para>Returns <c>true</c> if the system was shutdown during the specified time;
        /// <c>false</c> if it timed out, or the cancellationToken was canceled. </para>
        /// </summary>
        /// <param name="timeout">The timeout.</param>
        /// <param name="cancellationToken">A cancellation token that cancels the wait operation.</param>
        /// <returns>Returns <c>true</c> if the system was shutdown during the specified time;
        /// <c>false</c> if it timed out, or the cancellationToken was canceled. </returns>
        public abstract bool AwaitTermination(TimeSpan timeout, CancellationToken cancellationToken);


        public abstract void Stop(ActorRef actor);
        private bool _isDisposed; //Automatically initialized to false;

        //Destructor:
        //~ActorSystem() 
        //{
        //    // Finalizer calls Dispose(false)
        //    Dispose(false);
        //}

        /// <summary>Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.</summary>
        void IDisposable.Dispose()
        {
            Dispose(true);
            //Take this object off the finalization queue and prevent finalization code for this object
            //from executing a second time.
            GC.SuppressFinalize(this);
        }


        /// <summary>Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.</summary>
        /// <param name="disposing">if set to <c>true</c> the method has been called directly or indirectly by a 
        /// user's code. Managed and unmanaged resources will be disposed.<br />
        /// if set to <c>false</c> the method has been called by the runtime from inside the finalizer and only 
        /// unmanaged resources can be disposed.</param>
        private void Dispose(bool disposing)
        {
            // If disposing equals false, the method has been called by the
            // runtime from inside the finalizer and you should not reference
            // other objects. Only unmanaged resources can be disposed.

            try
            {
                //Make sure Dispose does not get called more than once, by checking the disposed field
                if(!_isDisposed)
                {
                    if(disposing)
                    {
                        Log.Debug("Disposing system");
                        Shutdown();
                    }
                    //Clean up unmanaged resources
                }
                _isDisposed = true;
            }
            finally
            {
                // base.dispose(disposing);
            }
        }


        public abstract object RegisterExtension(IExtensionId extension);

        public abstract ActorRef ActorOf(Props props, string name = null);
        
        public abstract ActorSelection ActorSelection(ActorPath actorPath);
        public abstract ActorSelection ActorSelection(string actorPath);

        /// <summary>
        /// Block and prevent the main application thread from exiting unless
        /// the actor system is shut down.
        /// </summary>
        [Obsolete("Use AwaitTermination instead")]
        public void WaitForShutdown()
        {
            AwaitTermination();
        }
    }
}