﻿using System.Collections.Generic;
using Akka.Actor;
using System;
using System.Threading.Tasks;
using System.Linq;
using Akka.Util.Internal;

namespace Akka.Routing
{
    internal class NoRoutee : Routee
    {
        public override void Send(object message, ActorRef sender)
        {
            if (sender is LocalActorRef)
            {
                sender.AsInstanceOf<LocalActorRef>().Provider.DeadLetters.Tell(message);
            }
        }
    }

    public class Routee
    {
        public static readonly Routee NoRoutee = new NoRoutee();

        public virtual void Send(object message, ActorRef sender)
        {
        }

        public virtual Task Ask(object message, TimeSpan? timeout)
        {
            return null;
        }


        public static implicit operator Routee(ActorRef actorRef)
        {
            return new ActorRefRoutee(actorRef);
        }
    }

    public class ActorRefRoutee : Routee
    {
        public ActorRef Actor { get; private set; }

        public ActorRefRoutee(ActorRef actor)
        {
            this.Actor = actor;
        }

        public override void Send(object message, ActorRef sender)
        {
            Actor.Tell(message, sender);
        }

        public override Task Ask(object message, TimeSpan? timeout)
        {
            return Actor.Ask(message, timeout);
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != this.GetType()) return false;
            return Equals((ActorRefRoutee)obj);
        }

        protected bool Equals(ActorRefRoutee other)
        {
            return Equals(Actor, other.Actor);
        }

        public override int GetHashCode()
        {
            return (Actor != null ? Actor.GetHashCode() : 0);
        }
    }

    public class ActorSelectionRoutee : Routee
    {
        private readonly ActorSelection _actor;

        public ActorSelection Selection { get { return _actor; } }

        public ActorSelectionRoutee(ActorSelection actor)
        {
            _actor = actor;
        }

        public override void Send(object message, ActorRef sender)
        {
            _actor.Tell(message, sender);
        }

        public override Task Ask(object message, TimeSpan? timeout)
        {
            return _actor.Ask(message, timeout);
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != this.GetType()) return false;
            return Equals((ActorSelectionRoutee)obj);
        }

        protected bool Equals(ActorSelectionRoutee other)
        {
            return Equals(_actor, other._actor);
        }

        public override int GetHashCode()
        {
            return (_actor != null ? _actor.GetHashCode() : 0);
        }
    }

    public class RouterEnvelope
    {
        public RouterEnvelope(object message)
        {
            Message = message;
        }

        public object Message { get; private set; }
    }

    public class Broadcast : RouterEnvelope
    {
        public Broadcast(object message)
            : base(message)
        {
        }
    }

    public class SeveralRoutees : Routee
    {
        private readonly Routee[] routees;

        public SeveralRoutees(Routee[] routees)
        {
            this.routees = routees;
        }

        public override void Send(object message, ActorRef sender)
        {
            foreach (Routee routee in  routees)
            {
                routee.Send(message, sender);
            }
        }
    }

    public abstract class RoutingLogic
    {
        public abstract Routee Select(object message, Routee[] routees);
    }

    public class Router
    {
        private readonly RoutingLogic _logic;
        private readonly Routee[] _routees;
       
        public Router(RoutingLogic logic, params Routee[] routees)
        {
            if(routees == null)
            {
                routees = new Routee[0];
            }
            _routees = routees;
            _logic = logic;
        }
        public IEnumerable<Routee> Routees
        {
            get { return _routees; }
        }

        public RoutingLogic RoutingLogic
        {
            get { return _logic; }
        }


        private object UnWrap(object message)
        {
            if (message is RouterEnvelope)
            {
                return message.AsInstanceOf<RouterEnvelope>().Message;
            }

            return message;
        }

        public void Route(object message, ActorRef sender)
        {
            if (message is Broadcast)
            {
                new SeveralRoutees(_routees).Send(UnWrap(message), sender);
            }
            else
            {
                Send(_logic.Select(message, _routees), message, sender);
            }
        }

        protected virtual void Send(Routee routee, object message, ActorRef sender)
        {
            routee.Send(UnWrap(message), sender);
        }

        /// <summary>
        /// Create a new instance with the specified routees and the same <see cref="RoutingLogic"/>.
        /// </summary>
        public Router WithRoutees(params Routee[] routees)
        {
            return new Router(_logic, routees);
        }

        /// <summary>
        /// Create a new instance with one more routee and the same <see cref="RoutingLogic"/>.
        /// </summary>
        public Router AddRoutee(Routee routee)
        {
            return new Router(_logic, _routees.Union(new[]{routee}).ToArray());
        }

        /// <summary>
        /// Create a new instance with one more routee and the same <see cref="RoutingLogic"/>.
        /// </summary>
        public Router AddRoutee(ActorRef routee)
        {
            return AddRoutee(new ActorRefRoutee(routee));
        }

        /// <summary>
        /// Create a new instance with one more routee and the same <see cref="RoutingLogic"/>.
        /// </summary>  
        public Router AddRoutee(ActorSelection routee)
        {
           return AddRoutee(new ActorSelectionRoutee(routee));
        }

        /// <summary>
        /// Create a new instance without the specified routee.
        /// </summary>
        public Router RemoveRoutee(Routee routee)
        {
            var routees = _routees.Where(r => !r.Equals(routee)).ToArray();
            return new Router(_logic, routees);
        }

        /// <summary>
        /// Create a new instance without the specified routee.
        /// </summary>
        public Router RemoveRoutee(ActorRef routee)
        {
            return RemoveRoutee(new ActorRefRoutee(routee));
        }

        /// <summary>
        /// Create a new instance without the specified routee.
        /// </summary>
        public Router RemoveRoutee(ActorSelection routee)
        {
            return RemoveRoutee(new ActorSelectionRoutee(routee));
        }
    }
}